#!/usr/bin/python

# Copyright 2010, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Author: Vojtech Aschenbrenner <vaschenb@redhat.com>

import sys
import operator
import autoqa.koji_utils
import autoqa.bodhi_utils
import rpmUtils.miscutils
from autoqa.repoinfo import repoinfo
from autoqa.test import AutoQATest
from autoqa.decorators import ExceptionCatcher
from autoqa.bodhi_utils import bodhi_post_testresult, query_update
from autoqa.koji_utils import strip_epoch

class upgradepath(AutoQATest):
    version = 1 # increment if setup() changes


    @ExceptionCatcher()
    def initialize(self, config, **kwargs):
        super(self.__class__, self).initialize(config)
        self.result = 'PASSED'
        # order for evaluation of final result; higher index means preference
        self.result_order = ('PASSED','INFO','FAILED','ABORTED')
        self.nvr_results = {} # results for invidual packages
        self.outputs = []
        self.highlights = []
        # dict of problematic builds (Koji/Bodhi problems): nvr -> error
        self.problematic = dict()

        self.koji = autoqa.koji_utils.SimpleKojiClientSession()
        self.koji.ensure_connection()


    def compare(self, proposed_build, tags, op):
        '''Compare proposed_build with given operator to latest build in
        all tags and return result.
        @param tags list of tags. tag is a string or tuple of strings. if it is
            a tuple, then the contents of corresponding repos will be united
            while searching for latest build
        '''
        result = 'PASSED'

        # convert 'op' to human language for error message clarity
        opstr = {operator.eq: 'equal to',
                 operator.ge: 'greater than or equal to',
                 operator.gt: 'greater than',
                 operator.le: 'less than or equal to',
                 operator.lt: 'less than',
                 operator.ne: 'not equal to'}
        assert op in opstr, 'Used unsupported operator: %s' % op

        # test the build against all proposed tags or set of tags
        for tag_tuple in tags:
            # convert everything to tuples
            if isinstance(tag_tuple, str):
                tag_tuple = (tag_tuple,)
            # now find the latest build in repositories
            latest_build = None
            for tag in tag_tuple:
                build = self.koji.latest_by_tag(tag, proposed_build['name'])
                if latest_build is None or (build is not None and
                    autoqa.koji_utils.compareEVR(build, latest_build) > 0):
                    latest_build = build
            # and make human readable string for this tag tuple
            tag_str = ' + '.join([tag for tag in tag_tuple])

            if latest_build is None:
                # no such package in repos means everything is fine
                msg ='{0:<7}{1}'.format('[ OK ]', tag_str)
                print msg
                self.outputs.append(msg)

                msg = "\tNo such package: %s" % proposed_build['name']
                print msg
                self.outputs.append(msg)
            else:
                res = autoqa.koji_utils.compareEVR(proposed_build, latest_build)

                if op(res, 0):
                    ok = True
                    msg ='{0:<7}{1}'.format('[ OK ]', tag_str)
                    print msg
                    self.outputs.append(msg)
                else:
                    ok = False
                    msg ='{0:<7}{1}'.format('[FAIL]', tag_str)
                    print msg
                    self.outputs.append(msg)
                msg = "\tLatest package: %s" % autoqa.koji_utils.getENVR(latest_build)
                print msg
                self.outputs.append(msg)
                if not ok:
                    msg = '\tError: Proposed package must be %s the latest package' % opstr[op]
                    print msg
                    self.outputs.append(msg)
                    result = 'FAILED'
        return result


    @ExceptionCatcher()
    def run_once(self, kojitag, hook, **kwargs):
        super(self.__class__, self).run_once()

        # Ideally upgradepath should check just the new updates. But since we don't
        # yet support test re-scheduling [1], we have to work around somehow [2].
        # So let's just run upgradepath for *all* updates requesting their move to
        # <kojitag>. It is a little inefficient, but if for some package the result
        # changes, we will report it correctly (and that's more important).
        # When test re-scheduling is supported, we can revert this test from whole
        # tag testing to just new updates testing.
        # [1] https://fedorahosted.org/autoqa/ticket/245
        # [2] https://fedorahosted.org/autoqa/ticket/246

        # THE UPGRADEPATH ALGORITHM:
        #
        # == Pushing to main repository ==
        # Pushing PKG to F(N)-main means:
        #   1. PKG in F(lower)-main <= PKG to push
        #   2. PKG in F(higher)-main >= PKG to push
        #
        # == Pushing to updates repository ==
        # Pushing PKG to F(N)-updates means:
        #   1. PKG in F(lower)-main <= PKG to push
        #   2. PKG in F(lower)-updates <= PKG to push
        #   3. PKG in F(higher)-main union F(higher)-updates => PKG to push
        #
        # Note: If PKG doesn't exist in REPO, it also satisfies any condition
        #
        # References:
        #   https://fedorahosted.org/autoqa/ticket/309
        #   https://fedorahosted.org/autoqa/ticket/277
        #
        # FIXME: Branched releases before "Final Change Deadline" use -updates
        # tag for pushing into -main repository. That breaks this algorithm.
        # See: https://fedorahosted.org/autoqa/ticket/277#comment:2
        # For proper functionality we need to detect such branched releases and
        # adjust the algorithm accordingly (run "Pushing to main repository" part
        # even if the tag does not correspond to it).

        # Get a list of all repos we monitor (currently not -testing)
        # FIXME - perhaps we should only query for 'active' repos
        repos = [repoinfo.getrepo(reponame) for reponame in repoinfo.repos()
                 if not reponame.endswith('-testing')]
        assert kojitag in [repo['tag'] for repo in repos], 'Requested unsupported kojitag: %s' % kojitag

        # are we pushing to main repo or to updates repo?
        # we recognize main repository when it has no parents
        push_to_main = not bool(repoinfo.getrepo_by_tag(kojitag)['parents'])

        if push_to_main:
            # we need to get list of main repositories for F(higher) and F(lower)
            low_repos = [repo for repo in repos if not repo['parents'] and repo['tag'] < kojitag]
            hi_repos = [repo for repo in repos if not repo['parents'] and repo['tag'] > kojitag]
        else: # pushing to update repo
            # for both F(lower) and F(higher) we are interested in main and update repos
            low_repos = [repo for repo in repos if repo['tag'] < kojitag]
            hi_repos = [repo for repo in repos if repo['tag'] > kojitag]
        reposorter = lambda repo : repo['name']
        low_repos = sorted(low_repos, key=reposorter)
        hi_repos = sorted(hi_repos, key=reposorter)

        # gather list of builds to test
        if hook == "post-bodhi-update-batch":
            # get the list of all builds requesting move to the kojitag
            sourcetag = kojitag + '-pending'
            print 'Koji tag to be tested: %s' % sourcetag
            updates = self.koji.listTagged(sourcetag)
            envrs = sorted([autoqa.koji_utils.getENVR(u) for u in updates])
            print 'Builds to be tested: %s' % ' '.join(envrs)
            update_id = sourcetag
        else: # hook is post-bodhi-update
            update_id = kwargs['name'] or kwargs['id']
            envrs = kwargs['nvrs']

        # for every build let's test it
        for envr in envrs:
            msg = '%s\n%s into %s\n%s' % (40*'=', envr, kojitag, 40*'=')
            print msg
            self.outputs.append(msg)

            # get all info about this current build
            if hook == "post-bodhi-update-batch":
                proposed_build = self.koji.getBuild(envr)
                if proposed_build is None:
                    msg = 'This build does not exist in Koji: %s' % envr
                    print >> sys.stderr, msg
                    self.highlights.append(msg)
                    self.problematic[envr] = msg
                    continue
            else: # hook is post-bodhi-update
                # don't query Koji, allow non-existing builds to be specified manually
                (name, version, release, epoch, arch) = rpmUtils.miscutils.splitFilename(envr + '.noarch')
                proposed_build = {
                    'name': name,
                    'version' : version,
                    'release' : release,
                    'epoch' : epoch,
                    'nvr' : envr,
                }

            # if we want to push an older build, warn about that and perform further
            # checking for the most recent one
            # see https://fedorahosted.org/autoqa/ticket/230
            latest_build = self.koji.latest_by_tag(kojitag, proposed_build['name'])
            if latest_build is not None:
                if autoqa.koji_utils.compareEVR(proposed_build, latest_build) <= 0:
                    msg = 'Warning: Same or newer build already exists in the same repo.\n' \
                          'All checks will be performed for: %s' \
                          % autoqa.koji_utils.getENVR(latest_build)
                    print >> sys.stderr, msg
                    self.outputs.append(msg)
                    self.highlights.append(msg)
                    proposed_build = latest_build

            if push_to_main:
                # compare with lower tags, so version has to be greater or equal
                low_tags = [repo['tag'] for repo in low_repos]
                result = self.compare(proposed_build, low_tags, operator.ge)
                # compare with higher tags, so version has to be lower or equal
                hi_tags = [repo['tag'] for repo in hi_repos]
                result2 = self.compare(proposed_build, hi_tags, operator.le)
            else: # pushing to update repository
                # compare with lower tags, so version has to be greater or equal
                low_tags = [repo['tag'] for repo in low_repos]
                result = self.compare(proposed_build, low_tags, operator.ge)
                # compare with higher tags, so version has to be lower or equal
                # we have to group corresponding main and update repos
                # let's define dict parent -> [child1, child2]
                # e.g. 'dist-f14' -> ['dist-f14-updates', 'dist-f14-updates-candidate']
                tag_family = {}
                for repo in hi_repos:
                    # this presumes all repos are sorted by name/tag
                    if not bool(repo['parents']):
                        tag_family[repo['tag']] = []
                    else:
                        main_parent = repoinfo.getrepo(repo['parents'][0])
                        if main_parent['tag'] in tag_family:
                            tag_family[main_parent['tag']].append(repo['tag'])
                        else:
                            # it might happen that there is no main_parent for
                            # a repo. e.g. we want to push to dist-f14 (so we don't
                            # check this tag), but we still want to check dist-f14-updates
                            # in this case just add it as a key
                            tag_family[repo['tag']] = []

                hi_tags = []
                for family in sorted(tag_family):
                    tag_tuple = [family]
                    tag_tuple.extend(tag_family[family])
                    hi_tags.append(tag_tuple)

                result2 = self.compare(proposed_build, hi_tags, operator.le)

            # compute pkg result
            if self.result_order.index(result2) > self.result_order.index(result):
                result = result2
            self.nvr_results[strip_epoch(envr)] = result

            msg = 'RESULT: %s' % result
            print msg
            self.outputs.append(msg)

            # update self.result
            if self.result_order.index(result) > self.result_order.index(self.result):
                self.result = result

        print # empty line

        # create summary like "1 PASSED, 2 FAILED, 3 INFO"
        summary = []
        for res in self.result_order:
            if res in self.nvr_results.values():
                count = len([k for k in self.nvr_results.keys() if self.nvr_results[k] == res])
                summary.append('%d %s' % (count, res))
        summary = ', '.join(summary)

        # print summary
        msg = 'SUMMARY: ' + summary
        print msg
        self.outputs.append(msg)
        self.highlights.append(msg)

        self.summary = '%s for %s' % (summary, update_id)

        # This code simply takes all nvrs (=builds), and creates a dictionary
        # of update metadata (taken from bodhi).
        # The variable 'updates' then looks like
        #
        # updates = {updatename : bodhi_metadata}

        updates = {}
        for envr in envrs:
            bodhi_info = query_update(strip_epoch(envr))
            if bodhi_info is None:
                msg = "Update containing %s not found in Bodhi" % envr
                print >> sys.stderr, msg
                self.highlights.append(msg)
                self.problematic[envr] = msg
                continue

            if bodhi_info['title'] not in updates.keys():
                updates[bodhi_info['title']] = bodhi_info

        # Because bodhi is working on top of updates we need to check whether
        # all nvrs (builds) from the respective update were accepted. If at least
        # one of the nvrs failed, the whole update failed.

        update_result = dict() # update_ title -> update_result
        for update in updates.values():
            update_result[update['title']] = 'PASSED'
            for build in update['builds']:
                if build['nvr'] not in self.nvr_results:
                    msg = "Build %s is part of update '%s', but it wasn't scheduled for testing" % \
                          (build['nvr'], update['title'])
                    print >> sys.stderr, msg
                    self.highlights.append(msg)
                    self.problematic[build['nvr']] = msg
                    update_result[update['title']] = 'ABORTED'
                    continue
                if self.result_order.index(self.nvr_results[build['nvr']]) > self.result_order.index(update_result[update['title']]):
                    update_result[update['title']] = self.nvr_results[build['nvr']]

        # Report all results to Bodhi
        for title, result in update_result.items():
            print "Posting to Bodhi: %s %s" % (title, result)
            bodhi_post_testresult(title, 'upgradepath', result, self.output_url, 'noarch', karma=0)

        # if some broken Bodhi updates found, end the test as NEEDS_INSPECTION
        if self.problematic:
            print >> sys.stderr, 'Broken updates found, setting result to NEEDS_INSPECTION'
            self.result = 'NEEDS_INSPECTION'

        if self.result != 'PASSED':
            links = ['https://fedoraproject.org/wiki/Packaging/NamingGuidelines#Minor_release_bumps_for_old_branches',\
                'https://fedoraproject.org/wiki/Packaging/NamingGuidelines#Package_Versioning']
            msg = 'Hint: See following links for resolving upgradepath-related failures: \n%s\n%s' % (links[1],links[0])
            self.highlights.append(msg)
